# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import date
from logging import getLogger

from pyramid.httpexceptions import HTTPMovedPermanently

from charmworld.jobs.ingest import (
    add_files,
)
from charmworld.models import (
    Charm,
    CharmSource,
    QADataSource,
    getfs,
)
from charmworld.testing import (
    factory,
    login_request,
    ViewTestBase,
    WebTestBase,
)
from charmworld.testing.html import (
    find_answer_field,
    parse,
)
from charmworld.utils import quote_key
from charmworld.views.api import API3
from charmworld.views.charms import (
    _charm_view,
    _reconcile_revision,
    charm_collection,
    config,
    distro_charm,
    distro_charm_json,
    find_charm,
    found_charm_collection,
    hook,
    interface,
    personal_collection,
    personal_config,
    personal_hook,
    personal_series_collection,
    quality_edit,
    series_collection,
)


DISTRO_CHARM_DATA = {
    'address': 'cs:precise/foo',
    'name': 'foo',
    'series': 'precise',
    'owner': 'bar',
    'maintainer': 'John Doe <jdoe@example.com>',
    'maintainers': '',
    'summary': 'Remote terminal classroom over https',
    'description': 'foobar',
    'store_url': 'cs:precise/foo-1',
    'bzr_branch': '',
    'promulgated': True,
    'branch_deleted': False,
    'categories': [],
    'subordinate': False,
    'requires': {'database': {'interface': 'mongodb'}},
    'provides': {'website': {'interface': 'https'}},
    'config': {
        'options': {
            'script-interval': {
                'default': 5,
                'type': 'int',
                'description': 'The interval between script runs.'
            }
        }
    },
    'downloads': 0,
    'downloads_in_past_7_days': 0,
    'downloads_in_past_30_days': 0,
    'downloads_in_past_half_year': 0,
    'files': {
        'readme': {
            'subdir': '',
            'filename': 'README',
        },
        'install': {
            'subdir': 'hooks',
            'filename': 'install',
        },
    },
    'tests': {},
    'test_results': {},
    'proof': {
        'w': [
            ' no README file',
            ' (install:31) - hook accesses EC2 metadata service'
            ' directly'
        ]
    },
    'date_created': '2012-01-01',
    'error': '',
    'last_change': {
        'committer': u'John Doe <jdoe@example.com>',
        'message': u'maintainer',
        'revno': 12,
        'created': 1343082725.06
    },
    'hash': None,
}


class TestCharms(ViewTestBase):

    def test_found_charm_collection(self):
        """found_charm_collection() returns a dict of charm objects."""
        _id, one = factory.makeCharm(self.db, name='bee', promulgated=True)
        _id, two = factory.makeCharm(self.db, name='awk', promulgated=True)
        _id, three = factory.makeCharm(self.db, name='cat', promulgated=False)
        query = {"series": 'precise'}
        sub_only = False
        # The only_promulgated kwarg limits the sorted charms.
        collection = found_charm_collection(
            self.db, dict(query), sub_only, only_promulgated=True)
        promulgated_names = [charm.name for charm in collection['charms']]
        expected = [two['name'], one['name']]
        self.assertEqual(expected, promulgated_names)
        self.assertIs(False, collection['sub_only'])
        # The only_promulgated kwarg defaults to False, getting all sorted
        # charms.
        collection = found_charm_collection(self.db, dict(query), sub_only)
        all_names = [charm.name for charm in collection['charms']]
        expected = [two['name'], one['name'], three['name']]
        self.assertEqual(expected, all_names)

    def test_charm_collection_basic(self):
        """Simple sanity check."""
        with self.assertRaises(HTTPMovedPermanently) as e:
            charm_collection(self.getRequest())
        self.assertIn('/solutions', e.exception.location)

    def test_personal_collection(self):
        request = self.getRequest()
        request.matchdict = {'owner': 'foo'}
        with self.assertRaises(HTTPMovedPermanently) as e:
            personal_collection(request)
        self.assertIn('/q/foo', e.exception.location)

    def test_series_collection(self):
        request = self.getRequest()
        request.matchdict = {'series': 'one'}
        with self.assertRaises(HTTPMovedPermanently) as e:
            series_collection(request)
        self.assertIn('/q/one', e.exception.location)

    def test_personal_series_collection(self):
        request = self.getRequest()
        request.matchdict = {'series': 'one', 'owner': 'me'}
        with self.assertRaises(HTTPMovedPermanently) as e:
            personal_series_collection(request)
        self.assertIn('/q/me/one', e.exception.location)

    def test_interface(self):
        request = self.getRequest()
        request.matchdict = {'interface': 'httpX'}
        with self.assertRaises(HTTPMovedPermanently) as e:
            interface(request)
        self.assertIn('/q/httpX', e.exception.location)

    def test_config(self):
        request = self.getRequest()
        request.matchdict = {'charm': 'foo', 'series': 'precise'}
        with self.assertRaises(HTTPMovedPermanently) as e:
            config(request)
        self.assertIn('/foo/precise#configuration', e.exception.location)

    def test_personal_config(self):
        request = self.getRequest()
        request.matchdict = {
            'charm': 'foo',
            'series': 'precise',
            'owner': 'bar'
        }
        with self.assertRaises(HTTPMovedPermanently) as e:
            personal_config(request)
        self.assertIn('/bar/foo/precise#configuration', e.exception.location)

    def make_charm_with_files(self, promulgated):
        ignore, charm = factory.makeCharm(
            self.db, owner='bar', series='precise',
            promulgated=promulgated, with_real_files=True)
        branch_dir, revision, path = self.use_context(factory.charm_branch())
        charm['branch_dir'] = branch_dir
        charm['store_data'] = {'digest': revision, 'revision': 1}
        fs = getfs(self.db)
        log = getLogger('charm.scan')
        add_files(fs, branch_dir, charm, log)
        self.db.charms.update({'_id': charm['_id']}, charm)
        return charm

    def test_hook(self):
        request = self.getRequest()
        request.matchdict = {
            'charm': 'sample_charm',
            'series': 'precise',
            'hook': 'install',
        }
        with self.assertRaises(HTTPMovedPermanently) as e:
            hook(request)
        self.assertIn('/sample_charm/precise', e.exception.location)

    def test_personal_hook(self):
        request = self.getRequest()
        request.matchdict = {
            'charm': 'sample_charm',
            'series': 'precise',
            'hook': 'install',
            'owner': 'bar',
        }
        with self.assertRaises(HTTPMovedPermanently) as e:
            personal_hook(request)
        self.assertIn('/bar/sample_charm/precise', e.exception.location)

    def test_distro_charm(self):
        request = self.getRequest()
        request.matchdict = {
            'charm': 'sample_charm',
            'series': 'precise',
        }
        with self.assertRaises(HTTPMovedPermanently):
            distro_charm(request)

    def test_distro_charm_json(self):
        ignore, charm = factory.makeCharm(
            self.db, name='foo', owner='bar', series='precise',
            description='foobar', promulgated=True,
            date_created=date(2012, 1, 1))
        request = self.getRequest()
        request.matchdict = {
            'charm': 'foo',
            'series': 'precise',
        }
        response = distro_charm_json(request)
        self.assertEqual(DISTRO_CHARM_DATA, response.json_body)

    def test_valid_others(self):
        charm = factory.makeCharm(self.db, files={})[1]
        other_charm = Charm(factory.makeCharm(
            self.db, series=charm['series'], name=charm['name'],
            owner='jrandom')[1])
        response = _charm_view(self.getRequest(), charm)
        self.assertEqual(1, len(response['others']))
        self.assertEqual(other_charm, response['others'][0])
        other_charm.store_data['errors'] = ['error']
        self.db.charms.save(other_charm._representation)
        response = _charm_view(self.getRequest(), charm)
        self.assertItemsEqual([], response['others'])

    def test_charm_icons(self):
        self.enable_routes()
        no_icon_charm = factory.makeCharm(self.db, files={})[1]
        response = _charm_view(self.getRequest(), no_icon_charm)
        self.assertEqual(None, response['icon_path'])

        files = {quote_key('icon.svg'): {'filename': 'icon.svg'}}
        icon_charm = factory.makeCharm(self.db, files=files)[1]
        response = _charm_view(self.getRequest(), icon_charm)
        expect = '/api/2/charm/%s/file/icon.svg' % API3._get_api_id(
            Charm(icon_charm))
        self.assertEqual(expect, response['icon_path'])


class TestFindCharm(ViewTestBase):

    def test_find_charm_selects_latest_revision(self):
        kwargs = dict(owner='foo', series='bar', name='baz')
        old_id, old_charm = factory.makeCharm(self.db, store_revision=1,
                                              **kwargs)
        new_id, new_charm = factory.makeCharm(self.db, store_revision=2,
                                              **kwargs)
        request = self.getRequest()
        request.matchdict.update(kwargs)
        request.matchdict['charm'] = kwargs['name']
        charm_data = find_charm(request)
        self.assertEqual(new_id, charm_data['_id'])


class CharmsViewHelpersTestCase(ViewTestBase):

    def test_reconcile_revision_with_dashed_name(self):
        # The charm name is reconstructed when a revision is wrongly matched.
        request = self.getRequest()
        request.matchdict = {
            'charm': 'charm',
            'revision': 'name',
        }
        _reconcile_revision(request)
        self.assertEqual('charm-name', request.matchdict['charm'])
        self.assertNotIn('revision', request.matchdict)

    def test_reconcile_revision_with_revision(self):
        # The charm name is reconstructed when a revision is wrongly matched.
        request = self.getRequest()
        request.matchdict = {
            'charm': 'charm',
            'revision': '1',
        }
        _reconcile_revision(request)
        self.assertEqual('charm', request.matchdict['charm'])
        self.assertEqual('1', request.matchdict['revision'])

    def test_reconcile_revision_with_head_revision(self):
        # The HEAD revision is dropped.
        request = self.getRequest()
        request.matchdict = {
            'charm': 'charm',
            'revision': 'HEAD',
        }
        _reconcile_revision(request)
        self.assertEqual('charm', request.matchdict['charm'])
        self.assertNotIn('revision', request.matchdict)
        request.matchdict = {
            'charm': 'charm',
            'revision': 'head',
        }
        _reconcile_revision(request)
        self.assertEqual('charm', request.matchdict['charm'])
        self.assertNotIn('revision', request.matchdict)


class CharmsWebTestCase(WebTestBase):

    def test_charm_collection_with_missing_data(self):
        """Charms with missing data are rendered."""
        response = self.app.get('/charms/precise', status=301)
        self.assertIn('/q/precise', response.body)

    def test_route_personal_charm_with_revision(self):
        # The route /~owner/series/charm-1 finds charms using owner,
        # series, charm name, and revision. The deploy instruction
        # does not include the revision.
        response = self.app.get('/~owner/series/charm-1', status=301)
        self.assertIn('/u/owner/charm/series/1', response.body)

    def test_route_personal_charm_with_dash_and_revision(self):
        # /~owner/series/charm-name-1 route find the charm.
        # This is a corner-case where the revision and charm name are
        # ambigous because they are munged.
        response = self.app.get('/~owner/series/charm-name-1', status=301)
        self.assertIn('/u/owner/charm-name/series/1', response.body)

    def test_route_personal_charm_with_head_revision(self):
        # the /~owner/series/charm-name-HEAD route finds the charm.
        # The juju-gui apache rewrite versionless charm urls to help the GUI.
        response = self.app.get('/~owner/series/charm-HEAD', status=301)
        self.assertIn('/u/owner/charm/series', response.body)
        self.assertNotIn('HEAD', response.body)

    def test_route_personal_charm_without_revision(self):
        # The route /~owner/series/charm route finds the charm using owner
        # series, and charm name. Revision is ignored. The deploy instruction
        # matches the route.
        response = self.app.get('/~owner/series/charm', status=301)
        self.assertIn('/u/owner/charm/series', response.body)

    def test_route_personal_charm_json_with_revision(self):
        # The route /~owner/series/charm-1/json finds charms using owner,
        # series, charm name, and revision. The address key does not contain
        # the revision.
        ignore, charm = factory.makeCharm(
            self.db, owner='owner', series='series', name='charm', files={})
        response = self.app.get('/~owner/series/charm-1/json', status=200)
        self.assertEqual('cs:~owner/series/charm', response.json['address'])

    def test_route_personal_charm_json_without_revision(self):
        # The route /~owner/series/charm/json finds the charm using owner
        # series, and charm name. Revision is ignored. The address key matches
        # the route
        ignore, charm = factory.makeCharm(
            self.db, owner='owner', series='series', name='charm', files={})
        response = self.app.get('/~owner/series/charm/json', status=200)
        self.assertEqual('cs:~owner/series/charm', response.json['address'])

    def test_route_promulgated_charm_with_revision(self):
        # The route /series/charm-1 route finds promulgated charms using
        # The series, charm name and version. The deploy instruction does not
        # include the version.
        response = self.app.get('/series/charm-1', status=301)
        self.assertIn('/charm/series/1', response.body)

    def test_route_promulgated_charm_with_head_revision(self):
        # The route /series/charm-HEAD finds promulgated charms using
        # series, charm name, and head revision. The revision is not included
        # in the deploy instruction.
        response = self.app.get('/series/charm-HEAD', status=301)
        self.assertNotIn('HEAD', response.body)
        self.assertIn('/charm/series', response.body)

    def test_route_promulgated_charm_without_revision(self):
        # The route /series/charm finds the promulgated charms using just
        # the series and charm name. The deploy instruction matches the route.
        self.app.get('/series/charm', status=301)

    def test_charm_short_url_route(self):
        ignore, charm_data = factory.makeCharm(
            self.db, promulgated=True, series='series', name='charm', files={})
        charm = Charm(charm_data)
        self.app.get(charm.short_url, status=301)


class TestQAForm(ViewTestBase):

    def setUp(self):
        """Need to load the question data."""
        super(TestQAForm, self).setUp()
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        resp = quality_edit(request)

        self.assertIn('form', resp)
        self.assertIn('label', resp['form'])

    def test_form_renders_existing_data(self):
        charm = Charm(factory.makeCharm(self.db, promulgated=True)[1])
        qa = {
            'reliable': {
                '4847e034bb0a55fcbc8a3380d6a3ab80': '1',
                'a60b40bce263c47559d32435a89756fe': '0',
                '7ff6cc997b141399bcdf2c338fb4031c': '1',
                '9a736d9ab5ca1e16ddd942e7a62b8820': '0',
                '51dceb2909262166283ac17f2af504e7': '1',
                '9659919b570c6ef9a25df7261375e6bb': '',
            },
        }
        source = QADataSource.from_db(self.db)
        source.save_qa_data(charm, qa)
        request = self.getRequest()
        request.matchdict = {
            'charm': charm.name,
            'series': charm.series,
        }
        resp = quality_edit(request)
        parsed = parse(resp['form'])
        one = find_answer_field(parsed, '4847e034bb0a55fcbc8a3380d6a3ab80',
                                '1')
        self.assertIsNot(None, one.attrib.get('checked'))

    def test_charm_saves_data(self):
        """Verify good data causes the view to store qa data.

        This also verifies our schema a little bit as the fields are turned
        from strings in POST to int's per the schema to match the original
        factory generated data.

        """
        charm = factory.get_charm_json(promulgated=True)
        self.use_index_client()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        post_data = factory.makeQAAssessment()
        # Add the submit button to trigger the right path through the view.
        post_data['submit'] = 'submit'
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }

        quality_edit(request)

        # Remove the submit button we added since it's not stored into the
        # document.
        del post_data['submit']

        # Now the stored QA data should match the data submitted.
        qa_data_source = QADataSource.from_db(self.db)
        found = qa_data_source.get_qa_data(Charm(charm))
        for key in post_data:
            # Ignore any form keys that start with __
            if key.startswith('__'):
                continue

            # Ignore the deform fields that make up the nested structure
            # for the qa questions.
            if key.startswith('deform'):
                continue

            self.assertEqual(
                post_data[key],
                found.charm_data[key])

            # The mapped/nested data needs to exist in the updated charm data.
            self.assertIn('reliable', found.charm_data)
        for stored_charm in charm_source._get_all(charm['_id']):
            self.assertNotIn('qa', stored_charm)


class TestQAFormFunc(WebTestBase):
    """Functional tests for the QA Form Views."""

    def test_form_403s_for_non_charmers(self):
        """The QA form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db)
            # The status force the check against the status code.
            resp = resp.goto(
                'charms/%s/%s/qa/edit' % (charm['series'], charm['name']),
                status=403)
            self.assertEqual(403, resp.status_code)

    def test_form_secured_from_users(self):
        """The qa form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=True)
            resp = resp.goto(
                'charms/%s/%s/qa/edit' % (charm['series'],
                charm['name']),
                status=200)
            self.assertEqual(resp.status_code, 200)
